/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <pwd.h>
#include <sys/types.h>
#include <utmp.h>
#include <string>
#include <set>
#include <iostream>

#include "config.h"
#include <pango/pango-font.h>
#include <gtk/gtkrc.h>

#ifdef DEBUG
#define DEBUG_BACKGROUND
#define DEBUG_USERSLIST
#endif

using std::string;
using std::set;
using std::cout;
using std::endl;

// Current user name
static string g_currentUserName;
// Current number of users
static int g_currentUsersCount = 0;
// Time the applet was started
time_t g_startTime = 0;
// Last time the alarm was triggered
time_t g_lastAlarm = 0;
// Maximum number of users
static int g_maxUsers;
// Ignore root and current users ?
static int g_ignoreRoot = 0;
static int g_ignoreCurrentUser = 0;
static int g_ignoredRoot = 0;
static int g_ignoredCurrentUser = 0;
// Ignore all users ?
static int g_ignoreAllUsers = 0;
// List of users to ignore
static set<string> g_usersToIgnoreList;
// Alarm command
static char *g_alarmCmd;
// Alarm period
static int g_alarmPeriod;
// Preferences box widgets
static GaiSS g_maxUserSpin = {
	1, 99, 1
};
static GaiSS g_alarmPeriodSpin = {
	5, 1440, 1
};
static GaiPI g_appletsPrefs[] = {
	{GAI_SPINBUTTON, "Critical number of users: ", &g_maxUsers, &g_maxUsers, &g_maxUserSpin},
	{GAI_CHECKBUTTON, "Ignore root", &g_ignoreRoot, &g_ignoreRoot, NULL},
	{GAI_CHECKBUTTON, "Ignore current user", &g_ignoreCurrentUser, &g_ignoreCurrentUser, NULL},
	{GAI_TEXTENTRY, "Alarm command: ", &g_alarmCmd, &g_alarmCmd, NULL},
	{GAI_SPINBUTTON, "Alarm period in seconds: ", &g_alarmPeriod, &g_alarmPeriod, &g_alarmPeriodSpin},
	{GAI_END, NULL}
};
// Disable alarms menu item
GtkMenuItem *g_alarmsMenuItem = NULL;

// Sets the applet background
void setBackground(bool firstCall)
{
	char usersNbr[64];

	string appletText = "Users: ";
	snprintf(usersNbr, 64, "%d", g_currentUsersCount);
	appletText += usersNbr;
	// Create a text object
#if 0
	GdkPixbuf *bgImage = gai_text_create(appletText.c_str(),
		"Courier", 10, GAI_TEXT_NORMAL, 0, 0, 0);
#else
	GdkPixbuf *bgImage = gai_text_create_simple(appletText.c_str(), 0, 0, 0);
#endif
	if (bgImage == NULL)
	{
		return;
	}

	// Create the background ?
	if (firstCall == true)
	{
		gai_background_from_gdkpixbuf(bgImage,
			GAI_BACKGROUND_MAX_SIZE_IMAGE);
		// Set flags
		gai_flags_set(GAI_FLAGS_NEVER_ROTATE);
#ifdef DEBUG_BACKGROUND
		cout << "setBackground: created background" << endl;
#endif
	}
	else
	{
		// No, update it
		gai_draw_bg(bgImage, 0, 0,
			gdk_pixbuf_get_width(bgImage),
			gdk_pixbuf_get_height(bgImage),
			0, 0);
		gai_draw_update_bg();
#ifdef DEBUG_BACKGROUND
		cout << "setBackground: updated background" << endl;
#endif
	}

	// Free the text object
	g_object_unref(bgImage);
}

// Updates the users list
void updateUsersList(void)
{
	struct utmp *u = NULL;
	int usersCount = 0;

#ifdef DEBUG_USERSLIST
	cout << "updateUsersList: current user is " << g_currentUserName << endl;
#endif
	// Rewind to the beginning of utmp
	setutent();
	// Read utmp
	while ((u = getutent()))
	{
		if (u->ut_type == USER_PROCESS)
		{
			// Is this user in the ignore list ?
			if (g_usersToIgnoreList.find(u->ut_user) == g_usersToIgnoreList.end())
			{
				// No, this user is not in the list
				if (g_ignoreAllUsers)
				{
					// OK, don't increment the number of users and insert this user in the list
					g_usersToIgnoreList.insert(u->ut_user);
#ifdef DEBUG_USERSLIST
					cout << "updateUsersList: added " << u->ut_user << " to ignore list !" << endl;
#endif
				}
				else
				{
					usersCount++;
#ifdef DEBUG_USERSLIST
					cout << "updateUsersList: " << u->ut_user << " not in the ignore list !" << endl;
#endif
				}
			}
		}
	}
	// Close utmp
	endutent();

	// Update the users count
	if (usersCount != g_currentUsersCount)
	{
		g_currentUsersCount = usersCount;
	}
	// Make sure this flag is reset
	g_ignoreAllUsers = 0;
#ifdef DEBUG_USERSLIST
	cout << "updateUsersList: system has " << usersCount << "/" << g_maxUsers << " users" << endl;
#endif

	time_t now = time(NULL);
	// If the maximum number of users has been reached, trigger the alarm.
	// Unless we have already done so less than a period ago or that the applet
	// was started less than 5 seconds ago.
	if ((g_currentUsersCount >= g_maxUsers) &&
		(g_lastAlarm + g_alarmPeriod <= now) &&
		(g_startTime + 5 <= now))
	{
		// Has an alarm command been defined ?
		if ((g_alarmCmd == NULL) ||
			(strlen(g_alarmCmd) == 0))
		{
#ifdef DEBUG_USERSLIST
			cout << "updateUsersList: popup alarm" << endl;
#endif
			// No, display a warning popup
			string warnMsg = "System currently has ";
			char usersNbr[64];
			snprintf(usersNbr, 64, "%d", g_currentUsersCount);
			warnMsg += usersNbr;
			warnMsg += " users logged in !";
			// Display a popup
			gai_display_error_continue(warnMsg.c_str());
		}
		else
		{
#ifdef DEBUG_USERSLIST
			cout << "updateUsersList: command alarm" << endl;
#endif
			// Yes, run the command
			// FIXME: there is now way to determine if this worked or not
			gai_exec(g_alarmCmd); 
		}

		// Update the last alarm time
		g_lastAlarm = time(NULL);

		// Reenable the disable alarms menu now
		if (g_alarmsMenuItem != NULL)
		{
			gtk_widget_set_sensitive(GTK_WIDGET(g_alarmsMenuItem), true);
#ifdef DEBUG_USERSLIST
			cout << "updateUsersList: reenabling menu" << endl;
#endif
		}
	}
#ifdef DEBUG_USERSLIST
	else cout << "updateUsersList: didn't ring alarm " << now << " " << g_startTime + 5 << " " << g_lastAlarm + 60 << " " << g_currentUsersCount << "/" << g_maxUsers << endl;
#endif
}

// Changes the users list when the state of the ignore options changes
void setIgnoreUsersList(void)
{
	if ((!g_ignoreRoot) &&
		(g_ignoredRoot))
	{
		set<string>::iterator iter = g_usersToIgnoreList.find("root");
		if (iter != g_usersToIgnoreList.end())
		{
			g_usersToIgnoreList.erase(iter);
		}
	}
	else if ((g_ignoreRoot) &&
		(!g_ignoredRoot))
	{
		g_usersToIgnoreList.insert("root");
	}
	if ((!g_ignoreCurrentUser) &&
		(g_ignoredCurrentUser))
	{
		set<string>::iterator iter = g_usersToIgnoreList.find(g_currentUserName);
		if (iter != g_usersToIgnoreList.end())
		{
			g_usersToIgnoreList.erase(iter);
		}
	}
	else if ((g_ignoreCurrentUser) &&
		(!g_ignoredCurrentUser))
	{
		g_usersToIgnoreList.insert(g_currentUserName);
	}
	g_ignoredRoot = g_ignoreRoot;
	g_ignoredCurrentUser = g_ignoreCurrentUser;
}

// Loads preferences, initializes applet
void loadAppletState(void)
{
	// Load settings
	// Maximum number of users
	g_maxUsers = gai_load_int_with_default("usermon/max_user", 3);
	// Ignore root and current user ?
	g_ignoreRoot = gai_load_int_with_default("usermon/ign_root", 1);
	g_ignoreCurrentUser = gai_load_int_with_default("usermon/ign_user", 1);
	// Don't ignore all users
	g_ignoreAllUsers = 0;
	// Alarm command
	g_alarmCmd = gai_load_string_with_default("usermon/alarm_cmd", "");
	// Alarm period
	g_alarmPeriod = gai_load_int_with_default("usermon/alarm_period", 5);

	// Get the current user name
	struct passwd *currentPwd = getpwuid(geteuid());
	if ((currentPwd != NULL) &&
		(currentPwd->pw_name != NULL))
	{
		g_currentUserName = currentPwd->pw_name;
	}

	// Set the users list
	setIgnoreUsersList();
}

// Preferences callback
void updateAppletState(int prefsChanged, gpointer d)
{
	// Save settings
	// Maximum number of users
	gai_save_int("usermon/max_user", g_maxUsers);
	// Ignore root and current users
	gai_save_int("usermon/ign_root", g_ignoreRoot);
	gai_save_int("usermon/ign_user", g_ignoreCurrentUser);
	// Alarm command
	gai_save_string("usermon/alarm_cmd", g_alarmCmd);
	// Alarm period
	gai_save_int("usermon/alarm_period", g_alarmPeriod);

	// Set the users list
	setIgnoreUsersList();

	// Reset the last alarm time
	g_lastAlarm = 0;
}

// Disables alarms menu callback
void disableAlarms(GtkMenuItem *menuItem, gpointer d)
{
	if (menuItem == NULL)
	{
		return;
	}

	// This will ensure that updateUsersList() ignores all users next time it runs
	g_ignoreAllUsers = 1;

	// Is this the first time this callback is called ?
	if (g_alarmsMenuItem == NULL)
	{
		// Yes, it is : set the global variable
		g_alarmsMenuItem = menuItem;
	}
	// Disable the menu until next time the alarm is sound
	gtk_widget_set_sensitive(GTK_WIDGET(g_alarmsMenuItem), false);
}

// Background callback
void updateBackground(void)
{
	static unsigned int callNum = 1;

	if (callNum == 10)
	{
		// Update the users list
		updateUsersList();
		callNum = 0;		
	}
	setBackground(false);
	callNum++;
}

int main(int argc, char **argv)
{
	// Initialize GAI
	gai_init2(&g_appletDefinition, &argc, &argv);

	// Load the applet settings
	loadAppletState();
	// Set the background
	setBackground(true);
	// Initialize the users list
	updateUsersList();

	// Set preferences
	gai_preferences2(APPLET_NAME " - Preferences",
		    g_appletsPrefs,
			"This monitors who is logged in and raises an alarm\n"
			"when the number of users reaches a limit.\n",
		    (GaiCallback1 *)updateAppletState,
		    NULL);

	// Get the current time
	g_startTime = time(NULL);

	// Add the disable alarms menu
	gai_menu_add("Ignore current users", NULL, GAI_MENU_NONE, (void*)disableAlarms, NULL);

	// Update the background twice per second
	gai_signal_on_update((GaiCallback0 *)updateBackground, 500, NULL);

	// Start the applet
	gai_start();

	return EXIT_SUCCESS;
}
